"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var acorn = __importStar(require("acorn"));
var acorn_bigint_1 = __importDefault(require("acorn-bigint"));
var acorn_dynamic_import_1 = __importDefault(require("acorn-dynamic-import"));
var acorn_import_meta_1 = __importDefault(require("acorn-import-meta"));
var GlobalScope_1 = __importDefault(require("./ast/scopes/GlobalScope"));
var EntityPathTracker_1 = require("./ast/utils/EntityPathTracker");
var Chunk_1 = __importStar(require("./Chunk"));
var Module_1 = __importStar(require("./Module"));
var ModuleLoader_1 = require("./ModuleLoader");
var assetHooks_1 = require("./utils/assetHooks");
var buildPhase_1 = require("./utils/buildPhase");
var chunkColouring_1 = require("./utils/chunkColouring");
var entryHashing_1 = require("./utils/entryHashing");
var executionOrder_1 = require("./utils/executionOrder");
var path_1 = require("./utils/path");
var pluginDriver_1 = require("./utils/pluginDriver");
var relativeId_1 = __importDefault(require("./utils/relativeId"));
var timers_1 = require("./utils/timers");
function makeOnwarn() {
    var warned = Object.create(null);
    return function (warning) {
        var str = warning.toString();
        if (str in warned)
            return;
        console.error(str);
        warned[str] = true;
    };
}
function normalizeEntryModules(entryModules) {
    if (typeof entryModules === 'string') {
        return [{ alias: null, unresolvedId: entryModules }];
    }
    if (Array.isArray(entryModules)) {
        return entryModules.map(function (unresolvedId) { return ({ alias: null, unresolvedId: unresolvedId }); });
    }
    return Object.keys(entryModules).map(function (alias) { return ({
        alias: alias,
        unresolvedId: entryModules[alias]
    }); });
}
var Graph = /** @class */ (function () {
    function Graph(options, watcher) {
        var _a;
        var _this = this;
        this.assetsById = new Map();
        this.curChunkIndex = 0;
        this.moduleById = new Map();
        this.needsTreeshakingPass = false;
        this.phase = buildPhase_1.BuildPhase.LOAD_AND_PARSE;
        this.watchFiles = Object.create(null);
        this.externalModules = [];
        this.modules = [];
        this.curChunkIndex = 0;
        this.deoptimizationTracker = new EntityPathTracker_1.EntityPathTracker();
        this.cachedModules = new Map();
        if (options.cache) {
            if (options.cache.modules)
                for (var _i = 0, _b = options.cache.modules; _i < _b.length; _i++) {
                    var module_1 = _b[_i];
                    this.cachedModules.set(module_1.id, module_1);
                }
        }
        if (options.cache !== false) {
            this.pluginCache = (options.cache && options.cache.plugins) || Object.create(null);
            // increment access counter
            for (var name_1 in this.pluginCache) {
                var cache = this.pluginCache[name_1];
                for (var _c = 0, _d = Object.keys(cache); _c < _d.length; _c++) {
                    var key = _d[_c];
                    cache[key][0]++;
                }
            }
        }
        this.preserveModules = options.preserveModules;
        this.cacheExpiry = options.experimentalCacheExpiry;
        this.treeshake = options.treeshake !== false;
        if (this.treeshake) {
            this.treeshakingOptions = options.treeshake
                ? {
                    annotations: options.treeshake.annotations !== false,
                    propertyReadSideEffects: options.treeshake.propertyReadSideEffects !== false,
                    pureExternalModules: options.treeshake.pureExternalModules
                }
                : { propertyReadSideEffects: true, annotations: true, pureExternalModules: false };
            if (this.treeshakingOptions.pureExternalModules === true) {
                this.isPureExternalModule = function () { return true; };
            }
            else if (typeof this.treeshakingOptions.pureExternalModules === 'function') {
                this.isPureExternalModule = this.treeshakingOptions.pureExternalModules;
            }
            else if (Array.isArray(this.treeshakingOptions.pureExternalModules)) {
                var pureExternalModules_1 = new Set(this.treeshakingOptions.pureExternalModules);
                this.isPureExternalModule = function (id) { return pureExternalModules_1.has(id); };
            }
            else {
                this.isPureExternalModule = function () { return false; };
            }
        }
        else {
            this.isPureExternalModule = function () { return false; };
        }
        this.contextParse = function (code, options) {
            if (options === void 0) { options = {}; }
            return _this.acornParser.parse(code, __assign(__assign(__assign({}, Module_1.defaultAcornOptions), options), _this.acornOptions));
        };
        this.pluginDriver = pluginDriver_1.createPluginDriver(this, options, this.pluginCache, watcher);
        if (watcher) {
            var handleChange_1 = function (id) { return _this.pluginDriver.hookSeqSync('watchChange', [id]); };
            watcher.on('change', handleChange_1);
            watcher.once('restart', function () {
                watcher.removeListener('change', handleChange_1);
            });
        }
        this.shimMissingExports = options.shimMissingExports;
        this.scope = new GlobalScope_1.default();
        this.context = String(options.context);
        var optionsModuleContext = options.moduleContext;
        if (typeof optionsModuleContext === 'function') {
            this.getModuleContext = function (id) { return optionsModuleContext(id) || _this.context; };
        }
        else if (typeof optionsModuleContext === 'object') {
            var moduleContext_1 = new Map();
            for (var key in optionsModuleContext) {
                moduleContext_1.set(path_1.resolve(key), optionsModuleContext[key]);
            }
            this.getModuleContext = function (id) { return moduleContext_1.get(id) || _this.context; };
        }
        else {
            this.getModuleContext = function () { return _this.context; };
        }
        this.onwarn = options.onwarn || makeOnwarn();
        this.acornOptions = options.acorn || {};
        var acornPluginsToInject = [];
        acornPluginsToInject.push(acorn_dynamic_import_1.default);
        acornPluginsToInject.push(acorn_import_meta_1.default);
        acornPluginsToInject.push(acorn_bigint_1.default);
        if (options.experimentalTopLevelAwait) {
            this.acornOptions.allowAwaitOutsideFunction = true;
        }
        var acornInjectPlugins = options.acornInjectPlugins;
        acornPluginsToInject.push.apply(acornPluginsToInject, (Array.isArray(acornInjectPlugins)
            ? acornInjectPlugins
            : acornInjectPlugins
                ? [acornInjectPlugins]
                : []));
        this.acornParser = (_a = acorn.Parser).extend.apply(_a, acornPluginsToInject);
        this.moduleLoader = new ModuleLoader_1.ModuleLoader(this, this.moduleById, this.pluginDriver, options.external, typeof options.manualChunks === 'function' && options.manualChunks);
    }
    Graph.prototype.build = function (entryModules, manualChunks, inlineDynamicImports) {
        // Phase 1 – discovery. We load the entry module and find which
        // modules it imports, and import those, until we have all
        // of the entry module's dependencies
        var _this = this;
        timers_1.timeStart('parse modules', 2);
        return Promise.all([
            this.moduleLoader.addEntryModules(normalizeEntryModules(entryModules), true),
            manualChunks &&
                typeof manualChunks === 'object' &&
                this.moduleLoader.addManualChunks(manualChunks)
        ]).then(function (_a) {
            var _b = _a[0], entryModules = _b.entryModules, manualChunkModulesByAlias = _b.manualChunkModulesByAlias;
            if (entryModules.length === 0) {
                throw new Error('You must supply options.input to rollup');
            }
            for (var _i = 0, _c = Array.from(_this.moduleById.values()); _i < _c.length; _i++) {
                var module_2 = _c[_i];
                if (module_2 instanceof Module_1.default) {
                    _this.modules.push(module_2);
                    _this.watchFiles[module_2.id] = true;
                }
                else {
                    _this.externalModules.push(module_2);
                }
            }
            timers_1.timeEnd('parse modules', 2);
            _this.phase = buildPhase_1.BuildPhase.ANALYSE;
            // Phase 2 - linking. We populate the module dependency links and
            // determine the topological execution order for the bundle
            timers_1.timeStart('analyse dependency graph', 2);
            _this.link(entryModules);
            timers_1.timeEnd('analyse dependency graph', 2);
            // Phase 3 – marking. We include all statements that should be included
            timers_1.timeStart('mark included statements', 2);
            if (inlineDynamicImports) {
                if (entryModules.length > 1) {
                    throw new Error('Internal Error: can only inline dynamic imports for single-file builds.');
                }
            }
            for (var _d = 0, entryModules_1 = entryModules; _d < entryModules_1.length; _d++) {
                var module_3 = entryModules_1[_d];
                module_3.includeAllExports();
            }
            _this.includeMarked(_this.modules);
            // check for unused external imports
            for (var _e = 0, _f = _this.externalModules; _e < _f.length; _e++) {
                var externalModule = _f[_e];
                externalModule.warnUnusedImports();
            }
            timers_1.timeEnd('mark included statements', 2);
            // Phase 4 – we construct the chunks, working out the optimal chunking using
            // entry point graph colouring, before generating the import and export facades
            timers_1.timeStart('generate chunks', 2);
            if (!_this.preserveModules && !inlineDynamicImports) {
                chunkColouring_1.assignChunkColouringHashes(entryModules, manualChunkModulesByAlias);
            }
            // TODO: there is one special edge case unhandled here and that is that any module
            //       exposed as an unresolvable export * (to a graph external export *,
            //       either as a namespace import reexported or top-level export *)
            //       should be made to be its own entry point module before chunking
            var chunks = [];
            if (_this.preserveModules) {
                for (var _g = 0, _h = _this.modules; _g < _h.length; _g++) {
                    var module_4 = _h[_g];
                    var chunk = new Chunk_1.default(_this, [module_4]);
                    if (module_4.isEntryPoint || !chunk.isEmpty) {
                        chunk.entryModules = [module_4];
                    }
                    chunks.push(chunk);
                }
            }
            else {
                var chunkModules = {};
                for (var _j = 0, _k = _this.modules; _j < _k.length; _j++) {
                    var module_5 = _k[_j];
                    var entryPointsHashStr = entryHashing_1.Uint8ArrayToHexString(module_5.entryPointsHash);
                    var curChunk = chunkModules[entryPointsHashStr];
                    if (curChunk) {
                        curChunk.push(module_5);
                    }
                    else {
                        chunkModules[entryPointsHashStr] = [module_5];
                    }
                }
                for (var entryHashSum in chunkModules) {
                    var chunkModulesOrdered = chunkModules[entryHashSum];
                    executionOrder_1.sortByExecutionOrder(chunkModulesOrdered);
                    var chunk = new Chunk_1.default(_this, chunkModulesOrdered);
                    chunks.push(chunk);
                }
            }
            // for each chunk module, set up its imports to other
            // chunks, if those variables are included after treeshaking
            for (var _l = 0, chunks_1 = chunks; _l < chunks_1.length; _l++) {
                var chunk = chunks_1[_l];
                chunk.link();
            }
            // filter out empty dependencies
            chunks = chunks.filter(Chunk_1.isChunkRendered);
            // then go over and ensure all entry chunks export their variables
            for (var _m = 0, chunks_2 = chunks; _m < chunks_2.length; _m++) {
                var chunk = chunks_2[_m];
                if (_this.preserveModules || chunk.entryModules.length > 0) {
                    chunk.generateEntryExportsOrMarkAsTainted();
                }
            }
            // create entry point facades for entry module chunks that have tainted exports
            var facades = [];
            if (!_this.preserveModules) {
                for (var _o = 0, chunks_3 = chunks; _o < chunks_3.length; _o++) {
                    var chunk = chunks_3[_o];
                    for (var _p = 0, _q = chunk.entryModules; _p < _q.length; _p++) {
                        var entryModule = _q[_p];
                        if (chunk.facadeModule !== entryModule) {
                            var entryPointFacade = new Chunk_1.default(_this, []);
                            entryPointFacade.turnIntoFacade(entryModule);
                            facades.push(entryPointFacade);
                        }
                    }
                }
            }
            timers_1.timeEnd('generate chunks', 2);
            _this.phase = buildPhase_1.BuildPhase.GENERATE;
            return chunks.concat(facades);
        });
    };
    Graph.prototype.finaliseAssets = function (assetFileNames) {
        var outputBundle = Object.create(null);
        this.assetsById.forEach(function (asset) {
            if (asset.source !== undefined)
                assetHooks_1.finaliseAsset(asset, outputBundle, assetFileNames);
        });
        return outputBundle;
    };
    Graph.prototype.getCache = function () {
        // handle plugin cache eviction
        for (var name_2 in this.pluginCache) {
            var cache = this.pluginCache[name_2];
            var allDeleted = true;
            for (var _i = 0, _a = Object.keys(cache); _i < _a.length; _i++) {
                var key = _a[_i];
                if (cache[key][0] >= this.cacheExpiry)
                    delete cache[key];
                else
                    allDeleted = false;
            }
            if (allDeleted)
                delete this.pluginCache[name_2];
        }
        return {
            modules: this.modules.map(function (module) { return module.toJSON(); }),
            plugins: this.pluginCache
        };
    };
    Graph.prototype.includeMarked = function (modules) {
        if (this.treeshake) {
            var treeshakingPass = 1;
            do {
                timers_1.timeStart("treeshaking pass " + treeshakingPass, 3);
                this.needsTreeshakingPass = false;
                for (var _i = 0, modules_1 = modules; _i < modules_1.length; _i++) {
                    var module_6 = modules_1[_i];
                    if (module_6.isExecuted)
                        module_6.include();
                }
                timers_1.timeEnd("treeshaking pass " + treeshakingPass++, 3);
            } while (this.needsTreeshakingPass);
        }
        else {
            // Necessary to properly replace namespace imports
            for (var _a = 0, modules_2 = modules; _a < modules_2.length; _a++) {
                var module_7 = modules_2[_a];
                module_7.includeAllInBundle();
            }
        }
    };
    Graph.prototype.warn = function (warning) {
        warning.toString = function () {
            var str = '';
            if (warning.plugin)
                str += "(" + warning.plugin + " plugin) ";
            if (warning.loc)
                str += relativeId_1.default(warning.loc.file) + " (" + warning.loc.line + ":" + warning.loc.column + ") ";
            str += warning.message;
            return str;
        };
        this.onwarn(warning);
    };
    Graph.prototype.link = function (entryModules) {
        for (var _i = 0, _a = this.modules; _i < _a.length; _i++) {
            var module_8 = _a[_i];
            module_8.linkDependencies();
        }
        var _b = executionOrder_1.analyseModuleExecution(entryModules), orderedModules = _b.orderedModules, cyclePaths = _b.cyclePaths;
        for (var _c = 0, cyclePaths_1 = cyclePaths; _c < cyclePaths_1.length; _c++) {
            var cyclePath = cyclePaths_1[_c];
            this.warn({
                code: 'CIRCULAR_DEPENDENCY',
                importer: cyclePath[0],
                message: "Circular dependency: " + cyclePath.join(' -> ')
            });
        }
        this.modules = orderedModules;
        for (var _d = 0, _e = this.modules; _d < _e.length; _d++) {
            var module_9 = _e[_d];
            module_9.bindReferences();
        }
        this.warnForMissingExports();
    };
    Graph.prototype.warnForMissingExports = function () {
        for (var _i = 0, _a = this.modules; _i < _a.length; _i++) {
            var module_10 = _a[_i];
            for (var _b = 0, _c = Object.keys(module_10.importDescriptions); _b < _c.length; _b++) {
                var importName = _c[_b];
                var importDescription = module_10.importDescriptions[importName];
                if (importDescription.name !== '*' &&
                    !importDescription.module.getVariableForExportName(importDescription.name)) {
                    module_10.warn({
                        code: 'NON_EXISTENT_EXPORT',
                        message: "Non-existent export '" + importDescription.name + "' is imported from " + relativeId_1.default(importDescription.module.id),
                        name: importDescription.name,
                        source: importDescription.module.id
                    }, importDescription.start);
                }
            }
        }
    };
    return Graph;
}());
exports.default = Graph;
