"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var ExternalModule_1 = __importDefault(require("./ExternalModule"));
var Module_1 = __importDefault(require("./Module"));
var error_1 = require("./utils/error");
var path_1 = require("./utils/path");
var referenceIds_1 = require("./utils/referenceIds");
var relativeId_1 = __importDefault(require("./utils/relativeId"));
var timers_1 = require("./utils/timers");
var transform_1 = __importDefault(require("./utils/transform"));
function normalizeRelativeExternalId(importer, source) {
    return path_1.isRelative(source) ? path_1.resolve(importer, '..', source) : source;
}
var ModuleLoader = /** @class */ (function () {
    function ModuleLoader(graph, modulesById, pluginDriver, external, getManualChunk) {
        var _this = this;
        this.entriesByReferenceId = new Map();
        this.entryModules = [];
        this.latestLoadModulesPromise = Promise.resolve();
        this.manualChunkModules = {};
        this.loadEntryModule = function (_a) {
            var alias = _a.alias, unresolvedId = _a.unresolvedId;
            return _this.pluginDriver
                .hookFirst('resolveId', [unresolvedId, undefined])
                .then(function (resolveIdResult) {
                if (resolveIdResult === false ||
                    (resolveIdResult && typeof resolveIdResult === 'object' && resolveIdResult.external)) {
                    error_1.error(error_1.errEntryCannotBeExternal(unresolvedId));
                }
                var id = resolveIdResult && typeof resolveIdResult === 'object'
                    ? resolveIdResult.id
                    : resolveIdResult;
                if (typeof id === 'string') {
                    return _this.fetchModule(id, undefined).then(function (module) {
                        if (alias !== null) {
                            if (module.chunkAlias !== null && module.chunkAlias !== alias) {
                                error_1.error(error_1.errCannotAssignModuleToChunk(module.id, alias, module.chunkAlias));
                            }
                            module.chunkAlias = alias;
                        }
                        return module;
                    });
                }
                error_1.error(error_1.errUnresolvedEntry(unresolvedId));
            });
        };
        this.graph = graph;
        this.modulesById = modulesById;
        this.pluginDriver = pluginDriver;
        if (typeof external === 'function') {
            this.isExternal = function (id, parentId, isResolved) {
                return !id.startsWith('\0') && external(id, parentId, isResolved);
            };
        }
        else {
            var ids_1 = new Set(Array.isArray(external) ? external : external ? [external] : []);
            this.isExternal = function (id) { return ids_1.has(id); };
        }
        this.getManualChunk = typeof getManualChunk === 'function' ? getManualChunk : function () { return null; };
    }
    ModuleLoader.prototype.addEntryModuleAndGetReferenceId = function (unresolvedEntryModule) {
        var entryRecord = {
            module: null,
            name: unresolvedEntryModule.unresolvedId
        };
        var referenceId = referenceIds_1.addWithNewReferenceId(entryRecord, this.entriesByReferenceId, unresolvedEntryModule.unresolvedId);
        this.addEntryModules([unresolvedEntryModule], false)
            .then(function (_a) {
            var module = _a.newEntryModules[0];
            entryRecord.module = module;
        })
            .catch(function () {
            // Avoid unhandled Promise rejection as the error will be thrown later
            // once module loading has finished
        });
        return referenceId;
    };
    ModuleLoader.prototype.addEntryModules = function (unresolvedEntryModules, isUserDefined) {
        var _this = this;
        var loadNewEntryModulesPromise = Promise.all(unresolvedEntryModules.map(this.loadEntryModule)).then(function (entryModules) {
            var _loop_1 = function (entryModule) {
                entryModule.isUserDefinedEntryPoint = entryModule.isUserDefinedEntryPoint || isUserDefined;
                var existingEntryModule = _this.entryModules.find(function (module) { return module.id === entryModule.id; });
                if (!existingEntryModule) {
                    _this.entryModules.push(entryModule);
                }
            };
            for (var _i = 0, entryModules_1 = entryModules; _i < entryModules_1.length; _i++) {
                var entryModule = entryModules_1[_i];
                _loop_1(entryModule);
            }
            return entryModules;
        });
        return this.awaitLoadModulesPromise(loadNewEntryModulesPromise).then(function (newEntryModules) { return ({
            entryModules: _this.entryModules,
            manualChunkModulesByAlias: _this.manualChunkModules,
            newEntryModules: newEntryModules
        }); });
    };
    ModuleLoader.prototype.addManualChunks = function (manualChunks) {
        var _this = this;
        var unresolvedManualChunks = [];
        for (var _i = 0, _a = Object.keys(manualChunks); _i < _a.length; _i++) {
            var alias = _a[_i];
            var manualChunkIds = manualChunks[alias];
            for (var _b = 0, manualChunkIds_1 = manualChunkIds; _b < manualChunkIds_1.length; _b++) {
                var unresolvedId = manualChunkIds_1[_b];
                unresolvedManualChunks.push({ alias: null, unresolvedId: unresolvedId, manualChunkAlias: alias });
            }
        }
        var loadNewManualChunkModulesPromise = Promise.all(unresolvedManualChunks.map(this.loadEntryModule)).then(function (manualChunkModules) {
            for (var index = 0; index < manualChunkModules.length; index++) {
                _this.addToManualChunk(unresolvedManualChunks[index].manualChunkAlias, manualChunkModules[index]);
            }
        });
        return this.awaitLoadModulesPromise(loadNewManualChunkModulesPromise);
    };
    ModuleLoader.prototype.getChunkFileName = function (referenceId) {
        var entryRecord = this.entriesByReferenceId.get(referenceId);
        if (!entryRecord)
            error_1.error(error_1.errChunkReferenceIdNotFoundForFilename(referenceId));
        var fileName = entryRecord.module &&
            (entryRecord.module.facadeChunk
                ? entryRecord.module.facadeChunk.id
                : entryRecord.module.chunk.id);
        if (!fileName)
            error_1.error(error_1.errChunkNotGeneratedForFileName(entryRecord));
        return fileName;
    };
    ModuleLoader.prototype.resolveId = function (source, importer) {
        var _this = this;
        return Promise.resolve(this.isExternal(source, importer, false)
            ? { id: source, external: true }
            : this.pluginDriver.hookFirst('resolveId', [source, importer])).then(function (result) { return _this.normalizeResolveIdResult(result, importer, source); });
    };
    ModuleLoader.prototype.addToManualChunk = function (alias, module) {
        if (module.manualChunkAlias !== null && module.manualChunkAlias !== alias) {
            error_1.error(error_1.errCannotAssignModuleToChunk(module.id, alias, module.manualChunkAlias));
        }
        module.manualChunkAlias = alias;
        if (!this.manualChunkModules[alias]) {
            this.manualChunkModules[alias] = [];
        }
        this.manualChunkModules[alias].push(module);
    };
    ModuleLoader.prototype.awaitLoadModulesPromise = function (loadNewModulesPromise) {
        var _this = this;
        this.latestLoadModulesPromise = Promise.all([
            loadNewModulesPromise,
            this.latestLoadModulesPromise
        ]);
        var getCombinedPromise = function () {
            var startingPromise = _this.latestLoadModulesPromise;
            return startingPromise.then(function () {
                if (_this.latestLoadModulesPromise !== startingPromise) {
                    return getCombinedPromise();
                }
            });
        };
        return getCombinedPromise().then(function () { return loadNewModulesPromise; });
    };
    ModuleLoader.prototype.fetchAllDependencies = function (module) {
        var _this = this;
        var fetchDynamicImportsPromise = Promise.all(module.getDynamicImportExpressions().map(function (specifier, index) {
            return _this.resolveDynamicImport(specifier, module.id).then(function (resolvedId) {
                if (resolvedId === null)
                    return;
                var dynamicImport = module.dynamicImports[index];
                if (typeof resolvedId === 'string') {
                    dynamicImport.resolution = resolvedId;
                    return;
                }
                return _this.fetchResolvedDependency(relativeId_1.default(resolvedId.id), module.id, resolvedId).then(function (module) {
                    dynamicImport.resolution = module;
                });
            });
        }));
        fetchDynamicImportsPromise.catch(function () { });
        return Promise.all(module.sources.map(function (source) { return _this.resolveAndFetchDependency(module, source); })).then(function () { return fetchDynamicImportsPromise; });
    };
    ModuleLoader.prototype.fetchModule = function (id, importer) {
        var _this = this;
        var existingModule = this.modulesById.get(id);
        if (existingModule) {
            if (existingModule.isExternal)
                throw new Error("Cannot fetch external module " + id);
            return Promise.resolve(existingModule);
        }
        var module = new Module_1.default(this.graph, id);
        this.modulesById.set(id, module);
        var manualChunkAlias = this.getManualChunk(id);
        if (typeof manualChunkAlias === 'string') {
            this.addToManualChunk(manualChunkAlias, module);
        }
        timers_1.timeStart('load modules', 3);
        return Promise.resolve(this.pluginDriver.hookFirst('load', [id]))
            .catch(function (err) {
            timers_1.timeEnd('load modules', 3);
            var msg = "Could not load " + id;
            if (importer)
                msg += " (imported by " + importer + ")";
            msg += ": " + err.message;
            throw new Error(msg);
        })
            .then(function (source) {
            timers_1.timeEnd('load modules', 3);
            if (typeof source === 'string')
                return source;
            if (source && typeof source === 'object' && typeof source.code === 'string')
                return source;
            error_1.error(error_1.errBadLoader(id));
        })
            .then(function (source) {
            var sourceDescription = typeof source === 'string'
                ? {
                    ast: null,
                    code: source
                }
                : source;
            var cachedModule = _this.graph.cachedModules.get(id);
            if (cachedModule &&
                !cachedModule.customTransformCache &&
                cachedModule.originalCode === sourceDescription.code) {
                // re-emit transform assets
                if (cachedModule.transformAssets) {
                    for (var _i = 0, _a = cachedModule.transformAssets; _i < _a.length; _i++) {
                        var asset = _a[_i];
                        _this.pluginDriver.emitAsset(asset.name, asset.source);
                    }
                }
                return cachedModule;
            }
            return transform_1.default(_this.graph, sourceDescription, module);
        })
            .then(function (source) {
            module.setSource(source);
            _this.modulesById.set(id, module);
            return _this.fetchAllDependencies(module).then(function () {
                for (var name_1 in module.exports) {
                    if (name_1 !== 'default') {
                        module.exportsAll[name_1] = module.id;
                    }
                }
                module.exportAllSources.forEach(function (source) {
                    var id = module.resolvedIds[source].id;
                    var exportAllModule = _this.modulesById.get(id);
                    if (exportAllModule instanceof ExternalModule_1.default)
                        return;
                    for (var name_2 in exportAllModule.exportsAll) {
                        if (name_2 in module.exportsAll) {
                            _this.graph.warn(error_1.errNamespaceConflict(name_2, module, exportAllModule));
                        }
                        else {
                            module.exportsAll[name_2] = exportAllModule.exportsAll[name_2];
                        }
                    }
                });
                return module;
            });
        });
    };
    ModuleLoader.prototype.fetchResolvedDependency = function (source, importer, resolvedId) {
        if (resolvedId.external) {
            if (!this.modulesById.has(resolvedId.id)) {
                this.modulesById.set(resolvedId.id, new ExternalModule_1.default({ graph: this.graph, id: resolvedId.id }));
            }
            var externalModule = this.modulesById.get(resolvedId.id);
            if (externalModule instanceof ExternalModule_1.default === false) {
                error_1.error(error_1.errInternalIdCannotBeExternal(source, importer));
            }
            return Promise.resolve(externalModule);
        }
        else {
            return this.fetchModule(resolvedId.id, importer);
        }
    };
    ModuleLoader.prototype.handleMissingImports = function (resolvedId, source, importer) {
        if (resolvedId === null) {
            if (path_1.isRelative(source)) {
                error_1.error(error_1.errUnresolvedImport(source, importer));
            }
            this.graph.warn(error_1.errUnresolvedImportTreatedAsExternal(source, importer));
            return { id: source, external: true };
        }
        return resolvedId;
    };
    ModuleLoader.prototype.normalizeResolveIdResult = function (resolveIdResult, importer, source) {
        var id = '';
        var external = false;
        if (resolveIdResult) {
            if (typeof resolveIdResult === 'object') {
                id = resolveIdResult.id;
                if (resolveIdResult.external) {
                    external = true;
                }
            }
            else {
                id = resolveIdResult;
                if (this.isExternal(id, importer, true)) {
                    external = true;
                }
            }
            if (external) {
                id = normalizeRelativeExternalId(importer, id);
            }
        }
        else {
            id = normalizeRelativeExternalId(importer, source);
            if (resolveIdResult !== false && !this.isExternal(id, importer, true)) {
                return null;
            }
            external = true;
        }
        return { id: id, external: external };
    };
    ModuleLoader.prototype.resolveAndFetchDependency = function (module, source) {
        var _this = this;
        return Promise.resolve(module.resolvedIds[source] ||
            this.resolveId(source, module.id).then(function (resolvedId) {
                return _this.handleMissingImports(resolvedId, source, module.id);
            })).then(function (resolvedId) {
            module.resolvedIds[source] = resolvedId;
            return _this.fetchResolvedDependency(source, module.id, resolvedId);
        });
    };
    ModuleLoader.prototype.resolveDynamicImport = function (specifier, importer) {
        var _this = this;
        // TODO we only should expose the acorn AST here
        return this.pluginDriver
            .hookFirst('resolveDynamicImport', [specifier, importer])
            .then(function (resolution) {
            if (typeof specifier !== 'string') {
                if (typeof resolution === 'string') {
                    return resolution;
                }
                if (!resolution) {
                    return null;
                }
                return __assign({ external: false }, resolution);
            }
            if (resolution == null) {
                return _this.resolveId(specifier, importer).then(function (resolvedId) {
                    return _this.handleMissingImports(resolvedId, specifier, importer);
                });
            }
            return _this.handleMissingImports(_this.normalizeResolveIdResult(resolution, importer, specifier), specifier, importer);
        });
    };
    return ModuleLoader;
}());
exports.ModuleLoader = ModuleLoader;
