"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = require("../../utils/path");
var MemberExpression_1 = __importDefault(require("./MemberExpression"));
var Node_1 = require("./shared/Node");
var ASSET_PREFIX = 'ROLLUP_ASSET_URL_';
var CHUNK_PREFIX = 'ROLLUP_CHUNK_URL_';
var MetaProperty = /** @class */ (function (_super) {
    __extends(MetaProperty, _super);
    function MetaProperty() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    MetaProperty.prototype.hasEffectsWhenAccessedAtPath = function (path) {
        return path.length > 1;
    };
    MetaProperty.prototype.initialise = function () {
        if (this.meta.name === 'import') {
            this.context.addImportMeta(this);
        }
        this.included = false;
    };
    MetaProperty.prototype.renderFinalMechanism = function (code, chunkId, format, pluginDriver) {
        if (!this.included)
            return false;
        var parent = this.parent;
        var importMetaProperty = parent instanceof MemberExpression_1.default && typeof parent.propertyKey === 'string'
            ? parent.propertyKey
            : null;
        if (importMetaProperty &&
            (importMetaProperty.startsWith(ASSET_PREFIX) || importMetaProperty.startsWith(CHUNK_PREFIX))) {
            var assetReferenceId = null;
            var chunkReferenceId = null;
            var fileName = void 0;
            if (importMetaProperty.startsWith(ASSET_PREFIX)) {
                assetReferenceId = importMetaProperty.substr(ASSET_PREFIX.length);
                fileName = this.context.getAssetFileName(assetReferenceId);
            }
            else {
                chunkReferenceId = importMetaProperty.substr(CHUNK_PREFIX.length);
                fileName = this.context.getChunkFileName(chunkReferenceId);
            }
            var relativePath = path_1.normalize(path_1.relative(path_1.dirname(chunkId), fileName));
            var replacement_1;
            if (assetReferenceId !== null) {
                // deprecated hook for assets
                replacement_1 = pluginDriver.hookFirstSync('resolveAssetUrl', [
                    {
                        assetFileName: fileName,
                        chunkId: chunkId,
                        format: format,
                        moduleId: this.context.module.id,
                        relativeAssetPath: relativePath
                    }
                ]);
            }
            if (!replacement_1) {
                replacement_1 = pluginDriver.hookFirstSync('resolveFileUrl', [
                    {
                        assetReferenceId: assetReferenceId,
                        chunkId: chunkId,
                        chunkReferenceId: chunkReferenceId,
                        fileName: fileName,
                        format: format,
                        moduleId: this.context.module.id,
                        relativePath: relativePath
                    }
                ]);
            }
            code.overwrite(parent.start, parent.end, replacement_1);
            return true;
        }
        var replacement = pluginDriver.hookFirstSync('resolveImportMeta', [
            importMetaProperty,
            {
                chunkId: chunkId,
                format: format,
                moduleId: this.context.module.id
            }
        ]);
        if (typeof replacement === 'string') {
            if (parent instanceof MemberExpression_1.default) {
                code.overwrite(parent.start, parent.end, replacement);
            }
            else {
                code.overwrite(this.start, this.end, replacement);
            }
            return true;
        }
        return false;
    };
    return MetaProperty;
}(Node_1.NodeBase));
exports.default = MetaProperty;
